@@ -1,4 +1,4 @@
-# $Id: ldiff.pl 29756 2010-01-04 16:34:34Z wsl $
+# $Id: ldiff.pl 32246 2010-08-23 13:51:54Z wsl $
 # $URL: https://infix.uvt.nl/its-id/trunk/sources/ldiff/script/ldiff.pl $
 
 =encoding utf8
@@ -9,7 +9,7 @@ C<ldiff> – calculate differences between LDIF files
 
 =head1 SYNOPSIS
 
- C<ldiff> I<file1> I<file2>
+C<ldiff> C<-[ih]> I<file1> I<file2>
 
 =head1 DESCRIPTION
 
@@ -19,6 +19,21 @@ change records to stdout that describe how to go from I<file1> to I<file2>.
 It determines the correct order to add and remove entries so that the LDAP
 tree is always consistent.
 
+=head1 USAGE
+
+=over
+
+=item C<-i>, C<--ignore>
+
+Ignore the (comma or whitespace separated) attributes while making
+the comparison.
+
+=item C<-i>, C<--ignore>
+
+Display a cheat sheet.
+
+=back
+
 =head1 EXAMPLE
 
  slapcat -l old.ldif
@@ -65,7 +80,7 @@ use warnings FATAL => 'all';
 use threads;
 
 use MIME::Base64;
-use IO::Handle;
+use IO::File;
 use Encode;
 use Getopt::Long qw(:config gnu_getopt);
 
@@ -74,41 +89,45 @@ my %ignore;
 sub canonical_entry {
 	my $line = $_[0];
 	my $off = index $line, ':';
-	die if $off == -1;
+	die 'internal error' if $off == -1;
 	my $key = substr $line, 0, $off;
 	my $type = substr $line, $off + 1, 1;
 
-	if($type eq ' ') {
+	if($type eq '') {
+		return (lc $key, '');
+	} elsif($type eq ' ') {
 		my $entry = substr $line, $off + 2;
-		return (lc $key, $entry);
+		return (lc $key, encode_utf8($entry));
 	} elsif($type eq ':') {
 		my $entry = substr $line, $off + 3;
 		return (lc $key, decode_base64($entry));
 	} else {
-		die "Don't know how to handle url type entries\n";
+		die "don't know how to handle '$type' type entries\n";
 	}
 }
 
 sub canonical_dn {
 	my $line = $_[0];
-	die if lc(substr($line, 0, 3)) ne 'dn:';
+	die 'internal error' if lc(substr($line, 0, 3)) ne 'dn:';
 	my $key = substr $line, 0, 2;
 	my $type = substr $line, 3, 1;
 
 	my $entry;
 
 	if($type eq ' ') {
-		$entry = substr $line, 4;
+		$entry = encode_utf8(substr($line, 4));
 	} elsif($type eq ':') {
-		$entry = decode_utf8(decode_base64(substr($line, 5)), Encode::FB_CROAK);
+		$entry = decode_base64(substr($line, 5));
+		decode_utf8($entry, Encode::FB_CROAK);
 	} else {
-		die "Don't know how to handle url type dn\n";
+		die "don't know how to handle '$type' type dn\n";
 	}
 
 	$entry =~ s/\s+/ /g;
 	$entry =~ s/^\s|\s$//g;
 	$entry =~ s/\s?,\s?/,/g;
 	$entry =~ s/,c=NL$/,c=NL/ig;
+
 	return $entry;
 }
 
@@ -124,13 +143,11 @@ sub dumpvalue {
 	$prefix = ''
 		unless defined $prefix;
 
-	if($val =~ /^[a-z0-9_]([ -~]*[a-z0-9_])?$/i) {
-		print "$prefix$key: $val\n"
-			or die;
+	if($val =~ /^[!-~]([ -~]*[!-~])?$/i) {
+		print "$prefix$key: $val\n" or die $!;
 	} else {
-		my $base64 = encode_base64(encode_utf8($val), '');
-		print "$prefix${key}:: $base64\n"
-			or die;
+		my $base64 = encode_base64($val, '');
+		print "$prefix${key}:: $base64\n" or die $!;
 	}
 }
 
@@ -143,18 +160,15 @@ sub dumprecord {
 	unless(defined $rec) {
 		return unless defined $ref;
 		dumpvalue('dn', $ref->{dn});
-		if(defined $modtype) {
-			print 'changetype: '.$modtype."\n"
-				or die;
-		}
+		print 'changetype: '.$modtype."\n" or die $!
+			if defined $modtype;
 		while(my ($key, $values) = each(%$ref)) {
 			next if $key eq 'dn';
 			foreach my $val (keys %$values) {
 				dumpvalue($key, $val, '# ');
 			}
 		}
-		print "\n"
-			or die;
+		print "\n" or die $!;
 		return;
 	}
 
@@ -162,7 +176,7 @@ sub dumprecord {
 
 	if(defined $modtype) {
 		print 'changetype: '.$modtype."\n"
-			or die;
+			or die $!;
 	}
 
 	while(my ($key, $values) = each(%$rec)) {
@@ -170,7 +184,7 @@ sub dumprecord {
 		if(defined $modify) {
 			next unless exists $modify->{$key};
 			print $modify->{$key}.": ".$key."\n"
-				or die;
+				or die $!;
 		}
 		if(exists $ref->{$key}) {
 			foreach my $val (keys %{$ref->{$key}}) {
@@ -181,7 +195,7 @@ sub dumprecord {
 			dumpvalue($key, $val);
 		}
 		if(defined $modify) {
-			print "-\n" or die;
+			print "-\n" or die $!;
 		}
 	}
 
@@ -189,20 +203,21 @@ sub dumprecord {
 		while(my ($key, $val) = each(%$modify)) {
 			next if exists $rec->{$key};
 			print "$val: $key\n"
-				or die;
+				or die $!;
 			if(exists $ref->{$key}) {
 				foreach my $val (keys %{$ref->{$key}}) {
 					dumpvalue($key, $val, '# ');
 				}
 			}
-			print "-\n" or die;
+			print "-\n" or die $!;
 		}
 	}
-	print "\n" or die;
+
+	print "\n" or die $!;
 }
 
 sub readldif {
-	my $name = shift;
+	my $file = shift;
 	my $entry = '';
 
 	my $dn;
@@ -215,46 +230,53 @@ sub readldif {
 		$records[3] = \%h;
 	}
 
-	open(my $fh, '<:utf8', $name)
-		or die "$name: $!\n";
+	my $fh = new IO::File($file, '<:utf8')
+		or die "$file: $!\n";
 
-	while(my $line = <$fh>) {
-		die "$name: $!\n" unless defined $line;
+	while(defined(my $line = $fh->getline)) {
 		chomp $line;
-		if($line eq '') {
-			if($entry ne '') {
-				if(defined $dn) {
-					my ($key, $val) = canonical_entry $entry;
-					undef $record->{$key}{$val}
-						unless exists $ignore{$key};
-				} else {
-					$dn = lc($record->{dn} = canonical_dn $entry);
+		eval {
+			if($line eq '') {
+				if($entry ne '') {
+					if(defined $dn) {
+						my ($key, $val) = canonical_entry $entry;
+						undef $record->{$key}{$val}
+							unless exists $ignore{$key};
+					} else {
+						$dn = lc($record->{dn} = canonical_dn $entry);
+					}
+					$entry = '';
 				}
-				$entry = '';
-			}
-			if(defined $dn) {
-				$records[level_dn $dn]{$dn} = $record;
-				undef $dn;
-				my %h;
-				keys %h = 10;
-				$record = \%h;
-			}
-		} elsif(ord($line) == 32) { # space
-			$entry .= substr($line, 1);
-		} elsif(ord($line) != 35) { # hash
-			if($entry ne '') {
 				if(defined $dn) {
-					my ($key, $val) = canonical_entry $entry;
-					undef $record->{$key}{$val}
-						unless exists $ignore{$key};
-				} else {
-					$dn = lc($record->{dn} = canonical_dn $entry);
+					$records[level_dn $dn]{$dn} = $record;
+					undef $dn;
+					my %h;
+					keys %h = 10;
+					$record = \%h;
 				}
+			} elsif(ord($line) == 32) { # space
+				$entry .= substr($line, 1);
+			} elsif(ord($line) != 35) { # hash
+				if($entry ne '') {
+					if(defined $dn) {
+						my ($key, $val) = canonical_entry $entry;
+						undef $record->{$key}{$val}
+							unless exists $ignore{$key};
+					} else {
+						$dn = lc($record->{dn} = canonical_dn $entry);
+					}
+				}
+				$entry = $line;
 			}
-			$entry = $line;
+		};
+		if($@) {
+			my $line = $fh->input_line_number;
+			die "$file:$line: $@";
 		}
 	}
-	die "$name: $!\n" if $fh->error;
+	$fh->eof or die "$file: $!\n";
+	$fh->close or die "$file: $!\n";
+	undef $fh;
 
 	if($entry ne '') {
 		if(defined $dn) {
@@ -272,8 +294,6 @@ sub readldif {
 		$record = {};
 	}
 
-	close($fh);
-
 	return \@records;
 }
 
@@ -342,14 +362,14 @@ sub deletes {
 }
 
 sub parse_ignore {
-	@ignore{split(/\s*,\s*/, lc $_[1])} = ();
+	@ignore{grep($_, split(/[,\s]+/, lc $_[1]))} = ();
 }
 
 sub usage {
 	my $fh = shift;
 	print $fh "Usage: findlinks [options] [dir [dir ...]]\n",
 		" -h, --help           Show usage information\n",
-		" -i, --ignore cn,uid  Comma-separated list of attrs to ignore\n";
+		" -i, --ignore cn,uid  List of attrs to ignore\n";
 }
 
 sub help {
@@ -365,10 +385,6 @@ unless(GetOptions(
 	exit 1;
 }
 
-foreach(keys %ignore) {
-	print STDERR "ignoring: $_\n"
-}
-
 my $draadje = threads->create(\&readldif, $ARGV[0]);
 
 #my $a = readldif($ARGV[0]);
